@@ -4,7 +4,7 @@
  * Copyright 2001 James Aylett
  * Copyright 2001,2002 Ananova Ltd
  * Copyright 2002 Intercede 1749 Ltd
- * Copyright 2002,2003,2004,2005,2006,2007,2008,2009,2010,2011,2013,2014,2015,2016 Olly Betts
+ * Copyright 2002,2003,2004,2005,2006,2007,2008,2009,2010,2011,2013,2014,2015,2016,2017,2018 Olly Betts
  * Copyright 2008 Thomas Viehmann
  *
  * This program is free software; you can redistribute it and/or
@@ -171,23 +171,31 @@ class MyStopper : public Xapian::Stopper {
 };
 
 static size_t
-prefix_from_term(string &prefix, const string &term)
+prefix_from_term(string* prefix, const string& term)
 {
-    if (term.empty()) {
-	prefix.resize(0);
-	return 0;
-    }
-    if (term[0] == 'X') {
-	const string::const_iterator begin = term.begin();
-	string::const_iterator i = begin + 1;
-	while (i != term.end() && C_isupper(*i)) ++i;
-	prefix.assign(begin, i);
-	if (i != term.end() && *i == ':') ++i;
-	return i - begin;
+    if (!term.empty()) {
+	if (term[0] == 'X') {
+	    const string::const_iterator begin = term.begin();
+	    string::const_iterator i = begin + 1;
+	    while (i != term.end() && C_isupper(*i))
+		++i;
+	    if (prefix)
+		prefix->assign(begin, i);
+	    if (i != term.end() && *i == ':')
+		++i;
+	    return i - begin;
+	}
+
+	if (C_isupper(term[0])) {
+	    if (prefix)
+		*prefix = term[0];
+	    return 1;
+	}
     }
 
-    prefix = term[0];
-    return 1;
+    if (prefix)
+	prefix->resize(0);
+    return 0;
 }
 
 // Don't allow ".." in format names, log file names, etc as this would allow
@@ -213,16 +221,16 @@ vet_filename(const string &filename)
 // BAD_QUERY parse error (message in error_msg)
 typedef enum { NEW_QUERY, SAME_QUERY, EXTENDED_QUERY, BAD_QUERY } querytype;
 
-static multimap<string, string> probabilistic_query;
+static multimap<string, string> query_strings;
 
 void
-set_probabilistic_query(const string & prefix, const string & s)
+add_query_string(const string& prefix, const string& s)
 {
     string query_string = s;
     // Strip leading and trailing whitespace from query_string.
     trim(query_string);
     if (!query_string.empty())
-	probabilistic_query.insert(make_pair(prefix, query_string));
+	query_strings.insert(make_pair(prefix, query_string));
 }
 
 static unsigned
@@ -313,7 +321,7 @@ read_qp_flags(const string & opt_pfx, unsigned f)
 }
 
 static querytype
-set_probabilistic(const string &oldp)
+parse_queries(const string& oldp)
 {
     // Parse the query string.
     qp.set_stemming_strategy(option["stem_all"] == "true" ? Xapian::QueryParser::STEM_ALL : Xapian::QueryParser::STEM_SOME);
@@ -355,13 +363,11 @@ set_probabilistic(const string &oldp)
 	    default_flags |= qp.FLAG_SPELLING_CORRECTION;
 
 	vector<Xapian::Query> queries;
-	queries.reserve(probabilistic_query.size());
+	queries.reserve(query_strings.size());
 
-	multimap<string, string>::const_iterator j;
-	for (j = probabilistic_query.begin();
-	     j != probabilistic_query.end();
-	     ++j) {
-	    const string & prefix = j->first;
+	for (auto& j : query_strings) {
+	    const string& prefix = j.first;
+	    const string& query_string = j.second;
 
 	    // Choose the stemmer to use for this input.
 	    string stemlang = option[prefix + ":stemmer"];
@@ -372,7 +378,6 @@ set_probabilistic(const string &oldp)
 	    // Work out the flags to use for this input.
 	    unsigned f = read_qp_flags(prefix + ":flag_", default_flags);
 
-	    const string & query_string = j->second;
 	    Xapian::Query q = qp.parse_query(query_string, f, prefix);
 	    if (!q.empty())
 		queries.push_back(q);
@@ -396,7 +401,7 @@ set_probabilistic(const string &oldp)
 
     // Check new query against the previous one
     if (oldp.empty()) {
-	// If oldp was empty that means there were no probabilistic terms
+	// If oldp was empty that means there were no parsed query terms
 	// before, so if there are now this is a new query.
 	return n_new_terms ? NEW_QUERY : SAME_QUERY;
     }
@@ -437,7 +442,7 @@ typedef multimap<string, string>::const_iterator FMCI;
 
 void add_bterm(const string &term) {
     string prefix;
-    if (prefix_from_term(prefix, term) > 0)
+    if (prefix_from_term(&prefix, term) > 0)
 	filter_map.insert(multimap<string, string>::value_type(prefix, term));
 }
 
@@ -473,8 +478,8 @@ run_query()
 			    op = Xapian::Query::OP_AND;
 			}
 			filter_vec.push_back(Xapian::Query(op,
-						     same_vec.begin(),
-						     same_vec.end()));
+							   same_vec.begin(),
+							   same_vec.end()));
 			break;
 		    }
 		}
@@ -489,7 +494,7 @@ run_query()
 			     filter_vec.begin(), filter_vec.end());
 
 	if (query.empty()) {
-	    // If no probabilistic query is provided then promote the filters
+	    // If no query strings were provided then promote the filters
 	    // to be THE query - filtering an empty query will give no
 	    // matches.
 	    std::swap(query, filter);
@@ -523,7 +528,7 @@ run_query()
 					Xapian::Query("Dlatest"));
 	}
 
-	// If no probabilistic query is provided then promote the daterange
+	// If no query strings were provided then promote the daterange
 	// filter to be THE query instead of filtering an empty query.
 	if (query.empty()) {
 	    query = date_filter;
@@ -911,6 +916,7 @@ CMD_cgilist,
 CMD_cgiparams,
 CMD_chr,
 CMD_collapsed,
+CMD_cond,
 CMD_contains,
 CMD_csv,
 CMD_date,
@@ -990,11 +996,16 @@ CMD_setmap,
 CMD_setrelevant,
 CMD_slice,
 CMD_snippet,
+CMD_sort,
 CMD_split,
 CMD_stoplist,
 CMD_sub,
+CMD_subdb,
+CMD_subid,
 CMD_substr,
 CMD_suggestion,
+CMD_switch,
+CMD_termprefix,
 CMD_terms,
 CMD_thispage,
 CMD_time,
@@ -1004,6 +1015,7 @@ CMD_transform,
 CMD_truncate,
 CMD_uniq,
 CMD_unpack,
+CMD_unprefix,
 CMD_unstem,
 CMD_upper,
 CMD_url,
@@ -1042,6 +1054,7 @@ T(cgilist,	   1, 1, N, 0), // return list of values for cgi parameter
 T(cgiparams,	   0, 0, N, 0), // return list of cgi parameter names
 T(chr,		   1, 1, N, 0), // return UTF-8 for given Unicode codepoint
 T(collapsed,	   0, 0, N, 0), // return number of hits collapsed into this
+T(cond,		   2, N, 0, 0), // return position of substring, or empty string
 T(contains,	   2, 2, N, 0), // return position of substring, or empty string
 T(csv,		   1, 2, N, 0), // CSV string escaping
 T(date,		   1, 2, N, 0), // convert time_t to strftime format
@@ -1122,11 +1135,16 @@ T(setmap,	   1, N, N, 0), // set map of option values
 T(setrelevant,	   0, 1, N, Q), // set rset
 T(slice,	   2, 2, N, 0), // slice a list using a second list
 T(snippet,	   1, 2, N, M), // generate snippet from text
+T(sort,		   1, 2, N, M), // alpha sort a list
 T(split,	   1, 2, N, 0), // split a string to give a list
 T(stoplist,	   0, 0, N, Q), // return list of stopped terms
 T(sub,		   2, 2, N, 0), // subtract
+T(subdb,	   0, 1, N, 0), // name of subdb docid is in
+T(subid,	   0, 1, N, 0), // docid in the subdb#
 T(substr,	   2, 3, N, 0), // substring
 T(suggestion,	   0, 0, N, Q), // misspelled word correction suggestion
+T(switch,	   3, N, 1, 0), // return position of substring, or empty string
+T(termprefix,	   1, 1, N, 0), // get any prefix from a term
 T(terms,	   0, 1, N, M), // list of matching terms
 T(thispage,	   0, 0, N, M), // page number of current page
 T(time,		   0, 0, N, M), // how long the match took (in seconds)
@@ -1138,8 +1156,9 @@ T(transform,	   3, 4, N, 0), // transform with a regexp
 T(truncate,	   2, 4, N, 0), // truncate after a word
 T(uniq,		   1, 1, N, 0), // removed duplicates from a sorted list
 T(unpack,	   1, 1, N, 0), // convert 4 byte big endian binary string to a number
-T(unstem,	   1, 1, N, Q), // return list of probabilistic terms from
-				// the query which stemmed to this term
+T(unprefix,	   1, 1, N, 0), // remove any prefix from a term
+T(unstem,	   1, 1, N, Q), // return list of terms from the parsed query
+				// which stemmed to this term
 T(upper,	   1, 1, N, 0), // convert string to upper case
 T(url,		   1, 1, N, 0), // url encode argument
 T(value,	   1, 2, N, 0), // return document value
@@ -1169,6 +1188,22 @@ write_all(int fd, const char * buf, size_t count)
     return 0;
 }
 
+static const vector<string>&
+get_subdbs()
+{
+    static vector<string> subdbs;
+    if (subdbs.empty()) {
+	size_t p = 0, q;
+	while (true) {
+	    q = dbname.find('/', p);
+	    subdbs.emplace_back(dbname, p, q - p);
+	    if (q == string::npos) break;
+	    p = q + 1;
+	}
+    }
+    return subdbs;
+}
+
 static string
 eval(const string &fmt, const vector<string> &param)
 {
@@ -1290,7 +1325,6 @@ eval(const string &fmt, const vector<string> &param)
 		break;
 	    case CMD_add: {
 		int total = 0;
-		vector<string>::const_iterator i;
 		for (auto&& arg : args)
 		    total += string_to_int(arg);
 		value = str(total);
@@ -1301,7 +1335,7 @@ eval(const string &fmt, const vector<string> &param)
 		break;
 	    case CMD_allterms: {
 		// list of all terms indexing document
-		int id = q0;
+		Xapian::docid id = q0;
 		if (!args.empty()) id = string_to_int(args[0]);
 		for (Xapian::TermIterator term = db.termlist_begin(id);
 		     term != db.termlist_end(id); ++term) {
@@ -1355,6 +1389,19 @@ eval(const string &fmt, const vector<string> &param)
 		value = str(collapsed);
 		break;
 	    }
+	    case CMD_cond:
+		for (size_t i = 0; i < args.size(); i += 2) {
+		    if (i == args.size() - 1) {
+			// Handle optional "else" value.
+			value = eval(args[i], param);
+			break;
+		    }
+		    if (!eval(args[i], param).empty()) {
+			value = eval(args[i + 1], param);
+			break;
+		    }
+		}
+		break;
 	    case CMD_contains: {
 		size_t pos = args[1].find(args[0]);
 		if (pos != string::npos) {
@@ -1534,15 +1581,27 @@ eval(const string &fmt, const vector<string> &param)
 	    case CMD_fmt:
 		value = fmtname;
 		break;
-	    case CMD_freq:
-		try {
-		    value = str(mset.get_termfreq(args[0]));
-		} catch (const Xapian::InvalidOperationError&) {
-		    // An MSet will raise this error if it's empty and not
-		    // associated with a search.
-		    value = str(db.get_termfreq(args[0]));
+	    case CMD_freq: {
+		const string& term = args[0];
+		Xapian::doccount termfreq = 0;
+		if (done_query) {
+		    try {
+			termfreq = mset.get_termfreq(term);
+		    } catch (const Xapian::InvalidOperationError&) {
+			// In 1.4.x and earlier, InvalidOperationError is
+			// thrown if the MSet is empty and not associated with
+			// an Enquire object.  In 1.5.0 and later, a termfreq
+			// of 0 is returned for this case.
+		    }
 		}
+		if (termfreq == 0) {
+		    // We want $freq to work before the match is run, and we
+		    // don't want using it to force the match to run.
+		    termfreq = db.get_termfreq(term);
+		}
+		value = str(termfreq);
 		break;
+	    }
 	    case CMD_ge:
 		if (string_to_int(args[0]) >= string_to_int(args[1]))
 		    value = "true";
@@ -1577,18 +1636,15 @@ eval(const string &fmt, const vector<string> &param)
 #if 0
 		url_query_string = "?DB=";
 		url_query_string += dbname;
-		multimap<string, string>::const_iterator j;
-		for (j = probabilistic_query.begin();
-		     j != probabilistic_query.end();
-		     ++j) {
-		    if (j->first.empty()) {
+		for (auto& j : query_strings) {
+		    if (j.first.empty()) {
 			url_query_string += "&P=";
 		    } else {
 			url_query_string += "&P."
-			url_query_string += j->first;
+			url_query_string += j.first;
 			url_query_string += '=';
 		    }
-		    const char *q = j->second.c_str();
+		    const char *q = j.second.c_str();
 		    int ch;
 		    while ((ch = *q++) != '\0') {
 			switch (ch) {
@@ -1895,16 +1951,9 @@ eval(const string &fmt, const vector<string> &param)
 	    case CMD_not:
 		if (args[0].empty()) value = "true";
 		break;
-	    case CMD_now: {
-		char buf[64];
-		my_snprintf(buf, sizeof(buf), "%lu",
-			    static_cast<unsigned long>(time(NULL)));
-		// MSVC's snprintf omits the zero byte if the string if
-		// sizeof(buf) long.
-		buf[sizeof(buf) - 1] = '\0';
-		value = buf;
+	    case CMD_now:
+		value = str(static_cast<unsigned long>(time(NULL)));
 		break;
-	    }
 	    case CMD_opt:
 		if (args.size() == 2) {
 		    value = option[args[0] + "," + args[1]];
@@ -1941,12 +1990,9 @@ eval(const string &fmt, const vector<string> &param)
 		url_prettify(value);
 		break;
 	    case CMD_query: {
-		pair<multimap<string, string>::const_iterator,
-		     multimap<string, string>::const_iterator> r;
-		r = probabilistic_query.equal_range(args.empty() ?
-						    string() : args[0]);
-		multimap<string, string>::const_iterator j;
-		for (j = r.first; j != r.second; ++j) {
+		auto r = query_strings.equal_range(args.empty() ?
+						   string() : args[0]);
+		for (auto j = r.first; j != r.second; ++j) {
 		    if (!value.empty()) value += '\t';
 		    const string & s = j->second;
 		    size_t start = 0, tab;
@@ -1976,14 +2022,14 @@ eval(const string &fmt, const vector<string> &param)
 		break;
 	    }
 	    case CMD_record: {
-		int id = q0;
+		Xapian::docid id = q0;
 		if (!args.empty()) id = string_to_int(args[0]);
 		value = db.get_document(id).get_data();
 		break;
 	    }
 	    case CMD_relevant: {
 		// document id if relevant; empty otherwise
-		int id = q0;
+		Xapian::docid id = q0;
 		if (!args.empty()) id = string_to_int(args[0]);
 		map<Xapian::docid, bool>::iterator i = ticked.find(id);
 		if (i != ticked.end()) {
@@ -2074,6 +2120,63 @@ eval(const string &fmt, const vector<string> &param)
 				     "<strong>", "</strong>", "...");
 		break;
 	    }
+	    case CMD_sort: {
+		const string &list = args[0];
+		if (list.empty()) break;
+		bool uniq = false;
+		bool rev = false;
+		if (args.size() > 1) {
+		    for (auto opt_ch : args[1]) {
+			switch (opt_ch) {
+			    case 'r':
+				rev = true;
+				break;
+			    case 'u':
+				uniq = true;
+				break;
+			    default:
+				throw string("Unknown $sort option: ") + opt_ch;
+			}
+		    }
+		}
+		vector<string> items;
+		string::size_type split = 0, split2;
+		do {
+		    split2 = list.find('\t', split);
+		    items.emplace_back(list, split, split2 - split);
+		    split = split2 + 1;
+		} while (split2 != string::npos);
+
+		if (!rev) {
+		    sort(items.begin(), items.end());
+		} else {
+		    sort(items.begin(), items.end(),
+			 [](const string& a, const string& b) {
+			     return a > b;
+			 });
+		}
+
+		value.reserve(list.size());
+		bool tab = false;
+		const string* prev = nullptr;
+		for (auto&& item : items) {
+		    // Skip duplicates if "u" flag specified.
+		    if (prev && *prev == item) {
+			continue;
+		    }
+		    if (uniq) {
+			prev = &item;
+		    }
+
+		    if (tab) {
+			value += '\t';
+		    } else {
+			tab = true;
+		    }
+		    value += item;
+		}
+		break;
+	    }
 	    case CMD_split: {
 		string split;
 		if (args.size() == 1) {
@@ -2110,6 +2213,19 @@ eval(const string &fmt, const vector<string> &param)
 	    case CMD_sub:
 		value = str(string_to_int(args[0]) - string_to_int(args[1]));
 		break;
+	    case CMD_subdb: {
+		Xapian::docid id = q0;
+		if (args.size() > 0) id = string_to_int(args[0]);
+		auto subdbs = get_subdbs();
+		value = subdbs[(id - 1) % subdbs.size()];
+		break;
+	    }
+	    case CMD_subid: {
+		Xapian::docid id = q0;
+		if (args.size() > 0) id = string_to_int(args[0]);
+		value = str(((id - 1) / get_subdbs().size()) + 1);
+		break;
+	    }
 	    case CMD_substr: {
 		int start = string_to_int(args[1]);
 		if (start < 0) {
@@ -2141,6 +2257,24 @@ eval(const string &fmt, const vector<string> &param)
 	    case CMD_suggestion:
 		value = qp.get_corrected_query_string();
 		break;
+	    case CMD_switch: {
+		const string& val = args[0];
+		for (size_t i = 1; i < args.size(); i += 2) {
+		    if (i == args.size() - 1) {
+			// Handle optional "else" value.
+			value = eval(args[i], param);
+			break;
+		    }
+		    if (val == eval(args[i], param)) {
+			value = eval(args[i + 1], param);
+			break;
+		    }
+		}
+		break;
+	    }
+	    case CMD_termprefix:
+		(void)prefix_from_term(&value, args[0]);
+		break;
 	    case CMD_terms: {
 		// list of matching terms
 		if (!enquire) break;
@@ -2264,6 +2398,11 @@ eval(const string &fmt, const vector<string> &param)
 	    case CMD_unpack:
 		value = str(binary_string_to_int(args[0]));
 		break;
+	    case CMD_unprefix: {
+		size_t prefix_len = prefix_from_term(NULL, args[0]);
+		value.assign(args[0], prefix_len, string::npos);
+		break;
+	    }
 	    case CMD_unstem: {
 		const string &term = args[0];
 		Xapian::TermIterator i = qp.unstem_begin(term);
@@ -2366,7 +2505,7 @@ pretty_term(string term)
 	// See if we have this prefix in the termprefix_to_userprefix map.  If
 	// so, just reverse the mapping (e.g. turn 'Sfish' into 'subject:fish').
 	string prefix;
-	size_t prefix_len = prefix_from_term(prefix, term);
+	size_t prefix_len = prefix_from_term(&prefix, term);
 
 	map<string, string>::const_iterator i;
 	i = termprefix_to_userprefix.find(prefix);
@@ -2459,7 +2598,7 @@ ensure_query_parsed()
 	discard_rset = true;
 	force_first_page = true;
     }
-    querytype result = set_probabilistic(v);
+    querytype result = parse_queries(v);
     switch (result) {
 	case BAD_QUERY:
 	    break;
@@ -2589,8 +2728,7 @@ OmegaExpandDecider::OmegaExpandDecider(const Xapian::Database & db_,
 	    }
 
 	    if (C_isupper(ch)) {
-		string prefix;
-		size_t prefix_len = prefix_from_term(prefix, term);
+		size_t prefix_len = prefix_from_term(NULL, term);
 		term.erase(0, prefix_len);
 	    }
 
